from collections import OrderedDict
import os
import socket
import sys
import time

import click
from flask import current_app

from . import server_backends
from .middleware import ReverseProxied
from .signals import (db_before_reset, db_reset_dropped, db_reset_created,
                      db_after_reset)
from .util import try_import_obj

ENV_DEFAULT = '.env'
APP_ENVVAR = 'FLASK_APP'


def register_cli(cli):
    @cli.command(help='Runs a development server with extras.')
    @click.option('--host',
                  '-h',
                  default='localhost',
                  help='Hostname to bind to. Defaults to localhost')
    @click.option('--port',
                  '-p',
                  type=int,
                  default=5000,
                  help='Port to listen on. Defaults to 5000')
    @click.option('--ssl',
                  '-S',
                  flag_value='adhoc',
                  default=None,
                  help='Enable SSL with a self-signed cert')
    @click.option('--gen-secret-key/--no-gen-secret-key',
                  default=True,
                  help='Enable or disable automatic secret key generation')
    @click.option('--flask-debug/--no-flask-debug',
                  '-e/-E',
                  default=None,
                  help='Enable/disable Flask-Debug or Flask-DebugToolbar '
                  'extensions. By default, both are enabled if debug is '
                  'enabled.')
    @click.option(
        '--extended-reload',
        '-R',
        default=2,
        type=float,
        help='Seconds before restarting the app if a non-recoverable '
        'exception occured (e.g. SyntaxError). Set this to 0 '
        'to disable (default: 2.0)')
    def dev(host, port, ssl, gen_secret_key, flask_debug, extended_reload):
        # FIXME: support all options of ``flask run``
        app = current_app

        if not app.debug:
            click.echo(' * app.debug = True')
            app.debug = True  # conveniently force debug mode
        extra_files = []

        if gen_secret_key and app.config.get('SECRET_KEY', None) is None:
            app.config['SECRET_KEY'] = os.urandom(64)

        # add configuration file to extra_files if passed in
        config_env_name = app.name.upper() + '_CONFIG'
        if config_env_name in os.environ:
            extra_files.append(os.environ[config_env_name])

        msgs = []

        # try to load debug extensions
        if flask_debug is None:
            flask_debug = app.debug

        if flask_debug:
            Debug = try_import_obj('flask_debug', 'Debug')
            DebugToolbarExtension = try_import_obj('flask_debugtoolbar',
                                                   'DebugToolbarExtension')

        if Debug:
            Debug(app)
            app.config['SERVER_NAME'] = '{}:{}'.format(host, port)

            # taking off the safety wheels
            app.config['FLASK_DEBUG_DISABLE_STRICT'] = True

        if DebugToolbarExtension:
            # Flask-Debugtoolbar does not check for debugging settings at
            # runtime. this hack enabled debugging if desired before
            # initializing the extension
            if app.debug:

                # set the SECRET_KEY, but only if we're in debug-mode
                if not app.config.get('SECRET_KEY', None):
                    msgs.append('SECRET_KEY not set, using insecure "devkey"')
                    app.config['SECRET_KEY'] = 'devkey'

            DebugToolbarExtension(app)

        def on_off(ext):
            return 'on' if ext is not None else 'off'

        msgs.insert(0, 'Flask-Debug: {}'.format(on_off(Debug)))
        msgs.insert(
            0, 'Flask-DebugToolbar: {}'.format(on_off(DebugToolbarExtension)))

        if msgs:
            click.echo(' * {}'.format(', '.join(msgs)))

        if extended_reload > 0:
            # we need to moneypatch the werkzeug reloader for this feature
            from werkzeug._reloader import ReloaderLoop
            orig_restart = ReloaderLoop.restart_with_reloader

            def _mp_restart(*args, **kwargs):
                while True:
                    status = orig_restart(*args, **kwargs)

                    if status == 0:
                        break
                    # an error occured, possibly a syntax or other
                    click.secho(
                        'App exited with exit code {}. Will attempted restart '
                        ' in {} seconds.'.format(status, extended_reload),
                        fg='red')
                    time.sleep(extended_reload)

                return status

            ReloaderLoop.restart_with_reloader = _mp_restart

        app.run(host, port, ssl_context=ssl, extra_files=extra_files)

    @cli.command(help='Runs a production server.')
    @click.option('--host',
                  '-H',
                  default='0.0.0.0',
                  help='Hostname to bind to. Defaults to 0.0.0.0')
    @click.option('--port',
                  '-p',
                  type=int,
                  default=80,
                  help='Port to listen on. Defaults to 80')
    @click.option('--processes',
                  '-w',
                  type=int,
                  default=1,
                  help='When possible, run this many instances in separate '
                  'processes. 0 means determine automatically. Default: 1')
    @click.option('--backends',
                  '-b',
                  default=server_backends.DEFAULT,
                  help='Comma-separated list of backends to try. Default: {}'
                  .format(server_backends.DEFAULT))
    @click.option(
        '--list',
        '-l',
        'list_only',
        is_flag=True,
        help='Do not run server, but list available backends for app.')
    @click.option(
        '--reverse-proxied',
        is_flag=True,
        help='Enable HTTP-reverse proxy middleware. Do not activate '
        'this unless you need it, it becomes a security risks when used '
        'incorrectly.')
    def serve(host, port, processes, backends, list_only, reverse_proxied):
        if processes <= 0:
            processes = None

        click.secho('flask serve is currently experimental. Use it at your '
                    'own risk',
                    fg='yellow',
                    err=True)
        app = current_app

        # we NEVER allow debug mode in production
        app.debug = False

        wsgi_app = app

        if reverse_proxied:
            wsgi_app = ReverseProxied(app)

        if list_only:
            found = False

            for backend in backends.split(','):
                try:
                    bnd = server_backends.backends[backend]
                except KeyError:
                    click.secho('{:20s} invalid'.format(backend), fg='red')
                    continue

                info = bnd.get_info()

                if info is None:
                    click.secho('{:20s} missing module'.format(backend),
                                fg='red')
                    continue

                fmt = {}
                if not found:
                    fmt['fg'] = 'green'
                    found = True

                click.secho(
                    '{b.name:20s} {i.version:10s} {i.extra_info}'.format(
                        b=bnd, i=info),
                    **fmt)
            return

        # regular operation
        for backend in backends.split(','):
            bnd = server_backends.backends[backend]
            info = bnd.get_info()
            if not info:
                continue

            b = bnd(processes)

            rcfg = OrderedDict()
            rcfg['app'] = app.name
            rcfg['# processes'] = str(b.processes)
            rcfg['backend'] = str(b)
            rcfg['addr'] = '{}:{}'.format(host, port)

            for k, v in rcfg.items():
                click.echo('{:15s}: {}'.format(k, v))

            try:
                b.run_server(wsgi_app, host, port)
                sys.exit(0)  # if the server exits normally, just quit
            except socket.error as e:
                if not port < 1024 or e.errno != 13:
                    raise

                # helpful message when trying to run on port 80 without room
                # permissions
                click.echo('Could not open socket on {}:{}: {}. '
                           'Do you have root permissions?'
                           .format(host, port, e))
                sys.exit(13)
            except RuntimeError as e:
                click.echo(str(e), err=True)
                sys.exit(1)
        else:
            click.echo('Exhausted list of possible backends', err=True)
            sys.exit(1)


def register_db_cli(cli, cli_mod):
    # FIXME: currently disabled
    @cli.group(help='Flask-SQLAlchemy functions')
    @click.option('--echo/--no-echo',
                  '-e/-E',
                  default=None,
                  help='Overrides SQLALCHEMY_ECHO')
    @cli_mod.with_appcontext
    def db(echo):
        # FIXME: currently broken.
        click.secho('flask db is currently experimental. Use it at your '
                    'own risk',
                    fg='yellow',
                    err=True)

        # sanity check
        if 'sqlalchemy' not in current_app.extensions:
            click.secho('No SQLAlchemy extension loaded. Did you initialize '
                        'your app?',
                        fg='red',
                        err=True)
            sys.exit(1)

        if echo is not None:
            current_app.config['SQLALCHEMY_ECHO'] = echo

    @db.command(help='Drop and recreated schema')
    def reset():
        app = current_app
        db = current_app.extensions['sqlalchemy'].db

        # FIXME: this should be in a transaction, but flask-sqlalchemy
        # currently makes it hard to get it right.
        #
        # problems that occured: con is not the same as the connection used
        # by drop_all and create_all, causing deadlocks to occur
        db_before_reset.send(app, db=db, con=db.engine)

        db.drop_all()
        db_reset_dropped.send(app, db=db, con=db.engine)

        db.create_all()
        db_reset_created.send(app, db=db, con=db.engine)

        db_after_reset.send(app, db=db, con=db.engine)



#!/usr/bin/env python

import os
from six.moves.urllib_parse import urlparse


def from_docker_envvars(config):
    # linked postgres database (link name 'pg' or 'postgres')
    if 'PG_PORT' in os.environ:
        pg_url = urlparse(os.environ['PG_PORT'])

        if not pg_url.scheme == 'tcp':
            raise ValueError('Only tcp scheme supported for postgres')

        host, port = pg_url.netloc.split(':')

        uri = 'postgres://{user}:{password}@{host}:{port}/{database}'.format(
            user=os.environ.get('PG_ENV_POSTGRES_USER', 'postgres'),
            password=os.environ.get('PG_ENV_POSTGRES_PASSWORD', ''),
            host=host,
            port=port,
            database=os.environ.get('PG_ENV_POSTGRES_DB', 'postgres'))

        config['SQLALCHEMY_DATABASE_URI'] = uri

    if 'REDIS_PORT' in os.environ:
        redis_url = urlparse(os.environ['REDIS_PORT'])

        if not redis_url.scheme == 'tcp':
            raise ValueError('Only tcp scheme supported for redis')

        host, port = redis_url.netloc.split(':')

        uri = 'redis://{host}:{port}/0'.format(host=host, port=port, )

        config['REDIS_URL'] = uri
        config['REDIS_HOST'] = host
        config['REDIS_PORT'] = int(port)



#!/usr/bin/env python

import os
import warnings

from . import env, heroku, docker
from .util import try_import


class AppConfig(object):
    def __init__(self, app=None, *args, **kwargs):
        if app:
            self.init_app(app, *args, **kwargs)

    def init_app(self,
                 app,
                 configfile=None,
                 envvar=True,
                 default_settings=True,
                 from_envvars='json',
                 from_envvars_prefix=None,
                 enable_cli=True):

        if from_envvars_prefix is None:
            from_envvars_prefix = app.name.upper() + '_'

        if default_settings is True:
            defs = try_import(app.name + '.default_config')
            if defs:
                app.config.from_object(defs)
        elif default_settings:
            app.config.from_object(default_settings)

        # load supplied configuration file
        if configfile is not None:
            warnings.warn('The configfile-parameter is deprecated and will '
                          'be removed (it is currently ignored). If you are '
                          'trying to load a configuration file, use the '
                          '_CONFIG envvar instead. During tests, simply '
                          'populate app.config before or after AppConfig.',
                          DeprecationWarning)

        # load configuration file from environment
        if envvar is True:
            envvar = app.name.upper() + '_CONFIG'

        if envvar and envvar in os.environ:
            app.config.from_envvar(envvar)

        # load environment variables
        if from_envvars:
            env.from_envvars(app.config,
                             from_envvars_prefix,
                             as_json=('json' == from_envvars))

        # register extension
        app.extensions = getattr(app, 'extensions', {})
        app.extensions['appconfig'] = self

        # register command-line functions if available
        if enable_cli:
            cli_mod = try_import('flask_cli', 'flask.cli')

            if hasattr(cli_mod, 'FlaskCLI') and not hasattr(app, 'cli'):
                # auto-load flask-cli if installed
                cli_mod.FlaskCLI(app)

            if hasattr(app, 'cli'):
                from .cli import register_cli, register_db_cli
                register_cli(app.cli)

                # conditionally register db api
                if try_import('flask_sqlalchemy'):
                    register_db_cli(app.cli, cli_mod)

        return app


class HerokuConfig(AppConfig):
    def init_app(self, app, *args, **kwargs):
        super(HerokuConfig, self).init_app(app, *args, **kwargs)

        heroku.from_heroku_envvars(app.config)


class DockerConfig(AppConfig):
    def init_app(self, app, *args, **kwargs):
        super(DockerConfig, self).init_app(app, *args, **kwargs)

        docker.from_docker_envvars(app.config)



from flask.signals import Namespace

# signals
signals = Namespace()
db_before_reset = signals.signal('db-before-reset')
db_reset_dropped = signals.signal('db-reset-dropped')
db_reset_created = signals.signal('db-reset-created')
db_after_reset = signals.signal('db-after-reset')



from collections import namedtuple
from multiprocessing import cpu_count

from .util import try_import


def _get_cpu_count():
    try:
        return cpu_count()
    except NotImplementedError:
        raise RuntimeError('Could not determine CPU count and no '
                           '--instance-count supplied.')


DEFAULT = 'tornado,meinheld,gunicorn,werkzeug-threaded,werkzeug'

backends = {}


def backend(name):
    def _(cls):
        backends[name] = cls
        cls.name = name
        return cls

    return _


BackendInfo = namedtuple('BackendInfo', 'version,extra_info')


class ServerBackend(object):
    vulnerable = True

    def __init__(self, processes=None):
        if not hasattr(self, 'processes'):
            if processes is None:
                processes = _get_cpu_count()
            self.processes = processes

    @classmethod
    def get_info(cls):
        """Return information about backend and its availability.

        :return: A BackendInfo tuple if the import worked, none otherwise.
        """
        mod = try_import(cls.mod_name)
        if not mod:
            return None
        version = getattr(mod, '__version__', None) or getattr(mod, 'version',
                                                               None)
        return BackendInfo(version or 'deprecated', '')

    def run_server(self, app, host, port):
        raise NotImplementedError

    def __str__(self):
        return '{} {}'.format(self.name, self.get_info().version)


@backend('werkzeug')
class WerkzeugBackend(ServerBackend):
    threaded = False
    mod_name = 'werkzeug'

    def run_server(self, app, host, port):
        app.run(host,
                port,
                debug=False,
                use_evalex=False,
                threaded=self.threaded,
                processes=self.processes)


@backend('werkzeug-threaded')
class WerkzeugThreaded(WerkzeugBackend):
    threaded = True
    processes = 1


@backend('tornado')
class TornadoBackend(ServerBackend):
    mod_name = 'tornado'

    def run_server(self, app, host, port):
        from tornado.wsgi import WSGIContainer
        from tornado.httpserver import HTTPServer
        from tornado.ioloop import IOLoop

        http_server = HTTPServer(WSGIContainer(app))
        http_server.listen(port, address=host)
        IOLoop.instance().start()


@backend('gunicorn')
class GUnicornBackend(ServerBackend):
    mod_name = 'gunicorn'

    def run_server(self, app, host, port):
        import gunicorn.app.base

        class FlaskGUnicornApp(gunicorn.app.base.BaseApplication):
            options = {
                'bind': '{}:{}'.format(host, port),
                'workers': self.processes
            }

            def load_config(self):
                for k, v in self.options.items():
                    self.cfg.set(k.lower(), v)

            def load(self):
                return app

        FlaskGUnicornApp().run()


@backend('meinheld')
class MeinHeldBackend(ServerBackend):
    mod_name = 'meinheld'

    def run_server(self, app, host, port):
        from meinheld import server

        server.listen((host, port))
        server.run(app)



from importlib import import_module


def try_import(*module_names):
    for module_name in module_names:
        try:
            return import_module(module_name)
        except ImportError:
            continue


def try_import_obj(module_name, name):
    mod = try_import(module_name)
    if mod:
        return getattr(mod, name, None)



#!/usr/bin/env python

import json
import os


def from_envvars(conf, prefix=None, envvars=None, as_json=True):
    """Load environment variables as Flask configuration settings.

    Values are parsed as JSON. If parsing fails with a ValueError,
    values are instead used as verbatim strings.

    :param app: App, whose configuration should be loaded from ENVVARs.
    :param prefix: If ``None`` is passed as envvars, all variables from
                   ``environ`` starting with this prefix are imported. The
                   prefix is stripped upon import.
    :param envvars: A dictionary of mappings of environment-variable-names
                    to Flask configuration names. If a list is passed
                    instead, names are mapped 1:1. If ``None``, see prefix
                    argument.
    :param as_json: If False, values will not be parsed as JSON first.
    """
    if prefix is None and envvars is None:
        raise RuntimeError('Must either give prefix or envvars argument')

    # if it's a list, convert to dict
    if isinstance(envvars, list):
        envvars = {k: None for k in envvars}

    if not envvars:
        envvars = {k: k[len(prefix):] for k in os.environ.keys()
                   if k.startswith(prefix)}

    for env_name, name in envvars.items():
        if name is None:
            name = env_name

        if not env_name in os.environ:
            continue

        if as_json:
            try:
                conf[name] = json.loads(os.environ[env_name])
            except ValueError:
                conf[name] = os.environ[env_name]
        else:
            conf[name] = os.environ[env_name]



#!/usr/bin/env python

import os
import re
import warnings
from six.moves.urllib_parse import urlparse

from . import env

HEROKU_POSTGRES_ENV_NAME_RE = re.compile('HEROKU_POSTGRESQL_[A-Z_]*URL')


def from_heroku_envvars(config):
    var_map = {
        # SQL-Alchemy
        'DATABASE_URL': 'SQLALCHEMY_DATABASE_URI',

        # Celery w/ RabbitMQ
        'BROKER_URL': 'RABBITMQ_URL',
        'REDISTOGO_URL': 'REDIS_URL',
        'MONGOLAB_URI': 'MONGO_URI',
        'MONGOHQ_URL': 'MONGO_URI',
        'CLOUDANT_URL': 'COUCHDB_URL',
        'MEMCACHIER_SERVERS': 'CACHE_MEMCACHED_SERVERS',
        'MEMCACHIER_USERNAME': 'CACHE_MEMCACHED_USERNAME',
        'MEMCACHIER_PASSWORD': 'CACHE_MEMCACHED_PASSWORD',
    }

    # search postgresql config using regex
    if 'DATABASE_URL' not in os.environ:
        for k in os.environ.keys():
            if HEROKU_POSTGRES_ENV_NAME_RE.match(k):
                var_map[k] = 'SQLALCHEMY_DATABASE_URI'
                warnings.warn('Using {0} as the database URL. However, '
                              'really should promote this or another URL '
                              'to DATABASE_URL by running \'heroku pg:'
                              'promote {0}\''.format(k), RuntimeWarning)

    var_list = [
        # Sentry
        'SENTRY_DSN',

        # Exceptional
        'EXCEPTIONAL_API_KEY',

        # Flask-GoogleFed
        'GOOGLE_DOMAIN',

        # Mailgun
        'MAILGUN_API_KEY',
        'MAILGUN_SMTP_LOGIN',
        'MAILGUN_SMTP_PASSWORD',
        'MAILGUN_SMTP_PORT',
        'MAILGUN_SMTP_SERVER',

        # SendGrid
        'SENDGRID_USERNAME',
        'SENDGRID_PASSWORD'
    ]

    # import the relevant envvars
    env.from_envvars(config, envvars=var_list, as_json=False)
    env.from_envvars(config, envvars=var_map, as_json=False)

    # fix up configuration
    if 'MAILGUN_SMTP_SERVER' in config:
        config['SMTP_SERVER'] = config['MAILGUN_SMTP_SERVER']
        config['SMTP_PORT'] = config['MAILGUN_SMTP_PORT']
        config['SMTP_LOGIN'] = config['MAILGUN_SMTP_LOGIN']
        config['SMTP_PASSWORD'] = config['MAILGUN_SMTP_PASSWORD']
        config['SMTP_USE_TLS'] = True
    elif 'SENDGRID_USERNAME' in config:
        config['SMTP_SERVER'] = 'smtp.sendgrid.net'
        config['SMTP_PORT'] = 25
        config['SMTP_LOGIN'] = config['SENDGRID_USERNAME']
        config['SMTP_PASSWORD'] = config['SENDGRID_PASSWORD']
        config['SMTP_USE_TLS'] = True

    # convert to Flask-Mail specific configuration
    if 'MAILGUN_SMTP_SERVER' in config or\
       'SENDGRID_PASSWORD' in config:

        config['MAIL_SERVER'] = config['SMTP_SERVER']
        config['MAIL_PORT'] = config['SMTP_PORT']
        config['MAIL_USE_TLS'] = config['SMTP_USE_TLS']
        config['MAIL_USERNAME'] = config['SMTP_LOGIN']
        config['MAIL_PASSWORD'] = config['SMTP_PASSWORD']

    # for backwards compatiblity, redis:
    if 'REDIS_URL' in config:
        url = urlparse(config['REDIS_URL'])
        config['REDIS_HOST'] = url.hostname
        config['REDIS_PORT'] = url.port
        config['REDIS_PASSWORD'] = url.password
        # FIXME: missing db#?

    if 'MONGO_URI' in config:
        url = urlparse(config['MONGO_URI'])
        config['MONGODB_USER'] = url.username
        config['MONGODB_PASSWORD'] = url.password
        config['MONGODB_HOST'] = url.hostname
        config['MONGODB_PORT'] = url.port
        config['MONGODB_DB'] = url.path[1:]
